def get_epoch_containing_block(block: int, netuid: int, tempo: int = 360) -> range:
    """
    Reimplementing the logic from subtensor's Rust function:
        pub fn blocks_until_next_epoch(netuid: u16, tempo: u16, block_number: u64) -> u64
    See https://github.com/opentensor/subtensor.

    See also: https://github.com/opentensor/bittensor/pull/2168/commits/9e8745447394669c03d9445373920f251630b6b8

    If given block happens to be an end of an epoch, the resulting epoch will end with it. The beginning of an epoch
    is the first block when values like "dividends" are different (before an epoch they are constant for a full
    tempo).
    """
    assert tempo > 0

    interval = tempo + 1
    last_epoch = block - 1 - (block + netuid + 1) % interval
    next_tempo_block_start = last_epoch + interval
    return range(last_epoch, next_tempo_block_start)


def get_cycle_containing_block(block: int, netuid: int, tempo: int = 360) -> range:
    """
    A cycle contains two epochs, starts on an even one. A cycle is the basic unit of passage of time in compute horde,
    and validators testing miners are synchronised to cycles.
    """
    very_first_epoch = get_epoch_containing_block(0, netuid, tempo=tempo)
    epoch_containing_block = get_epoch_containing_block(block, netuid, tempo=tempo)

    if ((epoch_containing_block.start - very_first_epoch.start) / (tempo + 1)) % 2:
        # that's the second epoch in this cycle
        first_epoch = range(
            epoch_containing_block.start - (tempo + 1), epoch_containing_block.stop - (tempo + 1)
        )
        second_epoch = epoch_containing_block
    else:
        first_epoch = epoch_containing_block
        second_epoch = range(
            epoch_containing_block.start + (tempo + 1), epoch_containing_block.stop + (tempo + 1)
        )

    return range(first_epoch.start, second_epoch.stop)



import base64
import typing
from typing import Annotated, Literal, Self

import pydantic
from pydantic import (
    BaseModel,
    JsonValue,
    field_serializer,
    field_validator,
    model_validator,
)

from compute_horde.base.output_upload import MultiUpload, OutputUpload, ZipAndHttpPutUpload
from compute_horde.base.volume import MultiVolume, Volume, ZipUrlVolume
from compute_horde.executor_class import ExecutorClass


class Error(BaseModel, extra="allow"):
    msg: str
    type: str
    help: str = ""


class Response(BaseModel, extra="forbid"):
    """Message sent from facilitator to validator in response to AuthenticationRequest & JobStatusUpdate"""

    status: Literal["error", "success"]
    errors: list[Error] = []


class Signature(BaseModel, extra="forbid"):
    # has defaults to allow easy instantiation
    signature_type: str = ""
    signatory: str = (
        ""  # identity of the signer (e.g. sa58 address if signature_type == "bittensor")
    )
    timestamp_ns: int = 0  # UNIX timestamp in nanoseconds
    signature: bytes

    @field_validator("signature")
    @classmethod
    def validate_signature(cls, signature: str) -> bytes:
        return base64.b64decode(signature)

    @field_serializer("signature")
    def serialize_signature(self, signature: bytes) -> str:
        return base64.b64encode(signature).decode("utf-8")


class V0JobRequest(BaseModel, extra="forbid"):
    """Message sent from facilitator to validator to request a job execution"""

    # this points to a `ValidatorConsumer.job_new` handler (fuck you django-channels!)
    type: Literal["job.new"] = "job.new"
    message_type: Literal["V0JobRequest"] = "V0JobRequest"

    uuid: str
    miner_hotkey: str
    executor_class: ExecutorClass
    docker_image: str
    raw_script: str
    args: list[str]
    env: dict[str, str]
    use_gpu: bool
    input_url: str
    output_url: str

    def get_args(self):
        return self.args

    @model_validator(mode="after")
    def validate_at_least_docker_image_or_raw_script(self) -> Self:
        if not (bool(self.docker_image) or bool(self.raw_script)):
            raise ValueError("Expected at least one of `docker_image` or `raw_script`")
        return self

    @property
    def volume(self) -> Volume | None:
        if self.input_url:
            return ZipUrlVolume(contents=self.input_url)
        return None

    @property
    def output_upload(self) -> OutputUpload | None:
        if self.output_url:
            return ZipAndHttpPutUpload(url=self.output_url)
        return None


class V1JobRequest(BaseModel, extra="forbid"):
    """Message sent from facilitator to validator to request a job execution"""

    # this points to a `ValidatorConsumer.job_new` handler (fuck you django-channels!)
    type: Literal["job.new"] = "job.new"
    message_type: Literal["V1JobRequest"] = "V1JobRequest"
    uuid: str
    miner_hotkey: str
    executor_class: ExecutorClass
    docker_image: str
    raw_script: str
    args: list[str]
    env: dict[str, str]
    use_gpu: bool
    volume: Volume | None = None
    output_upload: OutputUpload | None = None

    def get_args(self):
        return self.args

    @model_validator(mode="after")
    def validate_at_least_docker_image_or_raw_script(self) -> Self:
        if not (bool(self.docker_image) or bool(self.raw_script)):
            raise ValueError("Expected at least one of `docker_image` or `raw_script`")
        return self


class SignedFields(BaseModel):
    executor_class: str
    docker_image: str
    raw_script: str
    args: str
    env: dict[str, str]
    use_gpu: bool

    volumes: list[JsonValue]
    uploads: list[JsonValue]

    @staticmethod
    def from_facilitator_sdk_json(data: JsonValue):
        data = typing.cast(dict[str, JsonValue], data)

        signed_fields = SignedFields(
            executor_class=str(data.get("executor_class")),
            docker_image=str(data.get("docker_image", "")),
            raw_script=str(data.get("raw_script", "")),
            args=str(data.get("args", "")),
            env=typing.cast(dict[str, str], data.get("env", None)),
            use_gpu=typing.cast(bool, data.get("use_gpu")),
            volumes=typing.cast(list[JsonValue], data.get("volumes", [])),
            uploads=typing.cast(list[JsonValue], data.get("uploads", [])),
        )
        return signed_fields


def to_json_array(data) -> list[JsonValue]:
    return typing.cast(list[JsonValue], [x.model_dump() for x in data])


class V2JobRequest(BaseModel, extra="forbid"):
    """Message sent from facilitator to validator to request a job execution"""

    # this points to a `ValidatorConsumer.job_new` handler (fuck you django-channels!)
    type: Literal["job.new"] = "job.new"
    message_type: Literal["V2JobRequest"] = "V2JobRequest"
    signature: Signature | None = None

    uuid: str

    # !!! all fields below are included in the signed json payload
    executor_class: ExecutorClass
    docker_image: str
    raw_script: str
    args: list[str]
    env: dict[str, str]
    use_gpu: bool
    volume: Volume | None = None
    output_upload: OutputUpload | None = None
    # !!! all fields above are included in the signed json payload

    def get_args(self):
        return self.args

    def get_signed_fields(self) -> SignedFields:
        volumes = (
            to_json_array(
                self.volume.volumes if isinstance(self.volume, MultiVolume) else [self.volume]
            )
            if self.volume
            else []
        )

        uploads = (
            to_json_array(
                self.output_upload.uploads
                if isinstance(self.output_upload, MultiUpload)
                # TODO: fix consolidate faci output_upload types
                else [self.output_upload]  # type: ignore
            )
            if self.output_upload
            else []
        )

        signed_fields = SignedFields(
            executor_class=self.executor_class,
            docker_image=self.docker_image,
            raw_script=self.raw_script,
            args=" ".join(self.args),
            env=self.env,
            use_gpu=self.use_gpu,
            volumes=volumes,
            uploads=uploads,
        )
        return signed_fields

    def json_for_signing(self) -> JsonValue:
        payload = self.model_dump(mode="json")
        del payload["type"]
        del payload["message_type"]
        del payload["signature"]
        return payload

    @model_validator(mode="after")
    def validate_at_least_docker_image_or_raw_script(self) -> Self:
        if not (bool(self.docker_image) or bool(self.raw_script)):
            raise ValueError("Expected at least one of `docker_image` or `raw_script`")
        return self


JobRequest = Annotated[
    V0JobRequest | V1JobRequest | V2JobRequest,
    pydantic.Field(discriminator="message_type"),
]



from collections.abc import Iterable
from unittest.mock import MagicMock, patch

import pytest
from faker import Faker

from compute_horde.validators.tasks import fetch_validators
from tests.test_app.models import Validator


@pytest.fixture(autouse=True)
def mock_get_validators():
    with patch("compute_horde.validators.tasks.get_validators") as p:
        yield p


@pytest.fixture
def active_validators_keys(faker: Faker):
    return [faker.pystr() for _ in range(5)]


@pytest.fixture
def inactive_validators_keys(faker: Faker):
    return [faker.pystr() for _ in range(4)]


@pytest.fixture(autouse=True)
def active_validators(active_validators_keys: list[str]):
    objs = [Validator(public_key=key, active=True) for key in active_validators_keys]
    Validator.objects.bulk_create(objs)
    return objs


@pytest.fixture(autouse=True)
def inactive_validators(inactive_validators_keys: list[str]):
    objs = [Validator(public_key=key, active=False) for key in inactive_validators_keys]
    Validator.objects.bulk_create(objs)
    return objs


def _generate_validator_mocks(hotkeys: Iterable[str]):
    return [MagicMock(hotkey=key) for key in hotkeys]


@pytest.mark.django_db
def test_fetch_validators_creates_new_active_validators(
    faker: Faker, mock_get_validators: MagicMock
):
    new_keys = [faker.pystr() for _ in range(3)]
    mock_get_validators.return_value = _generate_validator_mocks(new_keys)

    fetch_validators()

    for key in new_keys:
        assert Validator.objects.filter(public_key=key, active=True).exists()


@pytest.mark.django_db
def test_fetch_validators_updates_existing_validators(
    inactive_validators_keys: list[str],
    active_validators: list[Validator],
    inactive_validators: list[Validator],
    mock_get_validators: MagicMock,
):
    mock_get_validators.return_value = _generate_validator_mocks(inactive_validators_keys)

    fetch_validators()

    assert Validator.objects.count() == len(active_validators) + len(inactive_validators)

    assert Validator.objects.filter(active=True).count() == len(inactive_validators)

    for key in inactive_validators_keys:
        assert Validator.objects.filter(public_key=key, active=True).exists()


@pytest.mark.django_db
def test_fetch_validators_debug(
    active_validators: list[Validator],
    active_validators_keys: list[str],
    inactive_validators: list[Validator],
    mock_get_validators: MagicMock,
    faker: Faker,
):
    mock_get_validators.return_value = _generate_validator_mocks(active_validators_keys)

    debug_validator_key = faker.pystr()
    Validator.objects.create(public_key=debug_validator_key, active=True, debug=True)

    fetch_validators()

    assert Validator.objects.count() == len(active_validators) + len(inactive_validators) + 1
    assert set(Validator.objects.filter(active=True).values_list("public_key", flat=True)) == set(
        active_validators_keys
    ) | {debug_validator_key}


@pytest.mark.django_db
def test_fetch_validators_debug_inactive(
    active_validators_keys: list[str],
    mock_get_validators: MagicMock,
    faker: Faker,
):
    mock_get_validators.return_value = _generate_validator_mocks(active_validators_keys)

    debug_validator_key = faker.pystr()
    debug_validator = Validator.objects.create(
        public_key=debug_validator_key, active=False, debug=True
    )

    fetch_validators()

    debug_validator.refresh_from_db()
    assert not debug_validator.active



from typing import Any, Literal, Self

import bittensor
from pydantic import BaseModel


class V0Heartbeat(BaseModel, extra="forbid"):
    """Message sent from validator to facilitator to keep connection alive"""

    message_type: Literal["V0Heartbeat"] = "V0Heartbeat"


class V0AuthenticationRequest(BaseModel, extra="forbid"):
    """Message sent from validator to facilitator to authenticate itself"""

    message_type: Literal["V0AuthenticationRequest"] = "V0AuthenticationRequest"
    public_key: str
    signature: str

    @classmethod
    def from_keypair(cls, keypair: bittensor.Keypair) -> Self:
        return cls(
            public_key=keypair.public_key.hex(),
            signature=f"0x{keypair.sign(keypair.public_key).hex()}",
        )

    def verify_signature(self) -> bool:
        public_key_bytes = bytes.fromhex(self.public_key)
        keypair = bittensor.Keypair(public_key=self.public_key, ss58_format=42)
        # make mypy happy
        valid: bool = keypair.verify(public_key_bytes, self.signature)
        return valid

    @property
    def ss58_address(self) -> str:
        # make mypy happy
        address: str = bittensor.Keypair(public_key=self.public_key, ss58_format=42).ss58_address
        return address


class V0MachineSpecsUpdate(BaseModel, extra="forbid"):
    """Message sent from validator to facilitator to update miner specs"""

    message_type: Literal["V0MachineSpecsUpdate"] = "V0MachineSpecsUpdate"
    miner_hotkey: str
    validator_hotkey: str
    specs: dict[str, Any]
    batch_id: str



from __future__ import annotations

import abc
import base64
import datetime
import hashlib
import json
import re
import time
import typing
from typing import ClassVar, Protocol

from class_registry import ClassRegistry, RegistryKeyError
from pydantic import JsonValue

from compute_horde.fv_protocol.facilitator_requests import Signature

if typing.TYPE_CHECKING:
    import bittensor

SIGNERS_REGISTRY: ClassRegistry[Signer] = ClassRegistry("signature_type")
VERIFIERS_REGISTRY: ClassRegistry[Verifier] = ClassRegistry("signature_type")


def verify_signature(
    payload: JsonValue | bytes,
    signature: Signature,
    *,
    newer_than: datetime.datetime | None = None,
):
    """
    Verifies the signature of the payload

    :param payload: payload to be verified
    :param signature: signature object
    :param newer_than: if provided, checks if the signature is newer than the provided timestamp
    :return: None
    :raises SignatureInvalidException: if the signature is invalid
    """
    try:
        verifier = VERIFIERS_REGISTRY.get(signature.signature_type)
    except RegistryKeyError as e:
        raise SignatureInvalidException(
            f"Invalid signature type: {signature.signature_type!r}"
        ) from e
    verifier.verify(payload, signature, newer_than)


class SignatureExtractor(Protocol):
    def __call__(self, headers: dict[str, str], prefix: str = "") -> Signature: ...


def signature_from_headers(headers: dict[str, str], prefix: str = "X-CH-") -> Signature:
    """
    Extracts the signature from the headers

    :param headers: headers dict
    :return: Signature object
    """
    try:
        return Signature(
            signature_type=headers[f"{prefix}Signature-Type"],
            signatory=headers[f"{prefix}Signatory"],
            timestamp_ns=int(headers[f"{prefix}Timestamp-NS"]),
            signature=headers[f"{prefix}Signature"].encode("utf-8"),
        )
    except (
        KeyError,
        ValueError,
        TypeError,
    ) as e:
        raise SignatureNotFound("Signature not found in headers") from e


def verify_request(
    method: str,
    url: str,
    headers: dict[str, str],
    json: JsonValue | None = None,
    *,
    newer_than: datetime.datetime | None = None,
    signature_extractor: SignatureExtractor = signature_from_headers,
) -> Signature | None:
    """
    Verifies the signature of the request

    :param method: HTTP method
    :param url: request URL
    :param headers: request headers
    :param json: request JSON payload
    :param newer_than: if provided, checks if the signature is newer than the provided timestamp
    :param signature_extractor: function to extract the signature from the headers
    :return: Signature object or None if no signature found
    :raises SignatureInvalidException: if the signature is invalid
    """
    try:
        signature = signature_extractor(headers)
    except SignatureNotFound:
        return None
    try:
        verifier = VERIFIERS_REGISTRY.get(signature.signature_type)
    except RegistryKeyError as e:
        raise SignatureInvalidException(
            f"Invalid signature type: {signature.signature_type!r}"
        ) from e
    payload = verifier.payload_from_request(method, url, headers=headers, json=json)
    verifier.verify(payload, signature, newer_than)
    return signature


def signature_to_headers(signature: Signature, prefix: str = "X-CH-") -> dict[str, str]:
    """
    Converts the signature to headers

    :param signature: Signature object
    :return: headers dict
    """
    return {
        f"{prefix}Signature-Type": signature.signature_type,
        f"{prefix}Signatory": signature.signatory,
        f"{prefix}Timestamp-NS": str(signature.timestamp_ns),
        f"{prefix}Signature": base64.b64encode(signature.signature).decode("utf-8"),
    }


class SignatureException(Exception):
    pass


class SignatureNotFound(SignatureException):
    pass


class SignatureInvalidException(SignatureException):
    pass


class SignatureTimeoutException(SignatureInvalidException):
    pass


def hash_message_signature(payload: bytes | JsonValue, signature: Signature) -> bytes:
    """
    Hashes the message to be signed with the signature parameters

    :param payload: payload to be signed
    :param signature: incomplete signature object with Signature parameters
    :return:
    """
    if not isinstance(payload, bytes):
        payload = json.dumps(payload, sort_keys=True).encode("utf-8")

    hasher = hashlib.blake2b()
    hasher.update(signature.timestamp_ns.to_bytes(8, "big"))
    hasher.update(payload)
    return hasher.digest()


_REMOVE_URL_SCHEME_N_HOST_RE = re.compile(r"^\w+://[^/]+")


def signature_payload(
    method: str, url: str, headers: dict[str, str], json: JsonValue | None = None
) -> JsonValue:
    reduced_url = _REMOVE_URL_SCHEME_N_HOST_RE.sub("", url)
    return {
        "action": f"{method.upper()} {reduced_url}",
        "json": json,
    }


class SignatureScheme(abc.ABC):
    signature_type: ClassVar[str]

    def payload_from_request(
        self,
        method: str,
        url: str,
        headers: dict[str, str],
        json: JsonValue | None = None,
    ):
        return signature_payload(
            method=method,
            url=url,
            headers=headers,
            json=json,
        )


class Signer(SignatureScheme):
    def sign(self, payload: JsonValue | bytes) -> Signature:
        signature = Signature(
            signature_type=self.signature_type,
            signatory=self.get_signatory(),
            timestamp_ns=time.time_ns(),
            signature=b"",
        )
        payload_hash = hash_message_signature(payload, signature)
        signature.signature = self._sign(payload_hash)
        return signature

    def signature_for_request(
        self, method: str, url: str, headers: dict[str, str], json: JsonValue | None = None
    ) -> Signature:
        return self.sign(self.payload_from_request(method, url, headers=headers, json=json))

    @abc.abstractmethod
    def _sign(self, payload: bytes) -> bytes:
        raise NotImplementedError

    @abc.abstractmethod
    def get_signatory(self) -> str:
        raise NotImplementedError


class Verifier(SignatureScheme):
    def verify(
        self,
        payload: JsonValue | bytes,
        signature: Signature,
        newer_than: datetime.datetime | None = None,
    ):
        payload_hash = hash_message_signature(payload, signature)
        self._verify(payload_hash, signature)

        if newer_than is not None:
            if newer_than > datetime.datetime.fromtimestamp(signature.timestamp_ns / 1_000_000_000):
                raise SignatureTimeoutException("Signature is too old")

    @abc.abstractmethod
    def _verify(self, payload: bytes, signature: Signature) -> None:
        raise NotImplementedError


def _require_bittensor():
    try:
        import bittensor
    except ImportError as e:
        raise ImportError("bittensor package is required for BittensorWalletSigner") from e
    return bittensor


class BittensorSignatureScheme:
    signature_type = "bittensor"


@SIGNERS_REGISTRY.register
class BittensorWalletSigner(BittensorSignatureScheme, Signer):
    def __init__(self, wallet: bittensor.wallet | bittensor.Keypair | None = None):
        bittensor = _require_bittensor()

        if isinstance(wallet, bittensor.Keypair):
            keypair = wallet
        else:
            keypair = (wallet or bittensor.wallet()).hotkey
        self._keypair = keypair

    def _sign(self, payload: bytes) -> bytes:
        signature: bytes = self._keypair.sign(payload)
        return signature

    def get_signatory(self) -> str:
        signatory: str = self._keypair.ss58_address
        return signatory


@VERIFIERS_REGISTRY.register
class BittensorWalletVerifier(BittensorSignatureScheme, Verifier):
    def __init__(self, *args, **kwargs):
        self._bittensor = _require_bittensor()

        super().__init__(*args, **kwargs)

    def _verify(self, payload: bytes, signature: Signature) -> None:
        try:
            keypair = self._bittensor.Keypair(ss58_address=signature.signatory)
        except ValueError:
            raise SignatureInvalidException("Invalid signatory for BittensorWalletVerifier")
        try:
            if not keypair.verify(data=payload, signature=signature.signature):
                raise SignatureInvalidException("Signature is invalid")
        except (ValueError, TypeError) as e:
            raise SignatureInvalidException("Signature is malformed") from e



import pytest

from compute_horde.subtensor import (
    get_cycle_containing_block,
    get_epoch_containing_block,
)


@pytest.mark.parametrize(
    ("netuid", "block", "expected_epoch"),
    [
        # netuid == 0
        (0, 25, range(-2, 359)),
        (0, 359, range(-2, 359)),
        (0, 360, range(359, 720)),
        (0, 720, range(359, 720)),
        (0, 721, range(720, 1081)),
        # netuid == 12
        (12, 25, range(-14, 347)),
        (12, 347, range(-14, 347)),
        (12, 348, range(347, 708)),
        (12, 708, range(347, 708)),
        (12, 709, range(708, 1069)),
        (12, 1100, range(1069, 1430)),
    ],
)
def test__get_epoch_containing_block(netuid, block, expected_epoch):
    assert (
        get_epoch_containing_block(block=block, netuid=netuid) == expected_epoch
    ), f"block: {block}, netuid: {netuid}, expected: {expected_epoch}"


@pytest.mark.parametrize(
    ("netuid", "block", "expected_cycle"),
    [
        # netuid == 0
        (0, 25, range(-2, 359 + 361)),
        (0, 359, range(-2, 359 + 361)),
        (0, 360, range(359 - 361, 720)),
        (0, 720, range(359 - 361, 720)),
        (0, 721, range(720, 1081 + 361)),
        # netuid == 12
        (12, 25, range(-14, 347 + 361)),
        (12, 347, range(-14, 347 + 361)),
        (12, 348, range(347 - 361, 708)),
        (12, 708, range(347 - 361, 708)),
        (12, 709, range(708, 1069 + 361)),
        (12, 1100, range(1069 - 361, 1430)),
    ],
)
def test__get_cycle_containing_block(netuid, block, expected_cycle):
    assert (
        get_cycle_containing_block(block=block, netuid=netuid) == expected_cycle
    ), f"block: {block}, netuid: {netuid}, expected: {expected_cycle}"



import asyncio
import ipaddress
import logging
import tempfile
from datetime import UTC, datetime, timedelta
from pathlib import Path

import aiohttp
from cryptography import x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
from cryptography.x509 import Certificate
from cryptography.x509.oid import NameOID

logger = logging.getLogger(__name__)


async def get_docker_container_ip(container_name: str, bridge_network: bool = False) -> str:
    query = (
        "'{{.NetworkSettings.Networks.bridge.IPAddress}}'"
        if bridge_network
        else "'{{range .NetworkSettings.Networks}}{{.IPAddress}}{{end}}'"
    )

    process = await asyncio.create_subprocess_exec(
        "docker",
        "inspect",
        "-f",
        query,
        container_name,
        stdout=asyncio.subprocess.PIPE,
    )
    stdout, _ = await process.communicate()
    await process.wait()
    if stdout:
        return stdout.decode().strip()[1:-1]
    else:
        raise Exception(f"Failed to get IP of {container_name}")


async def check_endpoint(url, timeout) -> bool:
    """
    Pings endpoint every second until it returns 200 or timeout is reached.
    """
    async with aiohttp.ClientSession() as session:
        for _ in range(timeout):
            try:
                response = await session.get(url)
                if response.status == 200:
                    return True
            except aiohttp.ClientError as e:
                logger.debug(f"Failed to ping {url}: {e}")
            await asyncio.sleep(1)
    return False


async def start_nginx(
    nginx_conf: str,
    port: int,
    dir_path: Path,
    job_network: str,
    container_name: str = "job-nginx",
    timeout: int = 10,
):
    nginx_conf_file = dir_path / "nginx.conf"
    nginx_conf_file.write_text(nginx_conf)

    cmd = [
        "docker",
        "run",
        "--detach",
        "--rm",
        "--name",
        container_name,
        "--network",
        "bridge",  # primary external network
        "-p",
        f"{port}:443",  # expose nginx port
        "-v",
        f"{dir_path}:/etc/nginx/",
        "nginx:1.26-alpine",
    ]
    process = await asyncio.create_subprocess_exec(*cmd)
    _stdout, _stderr = await process.communicate()
    await process.wait()

    # make sure to get ip while there is still a single network
    ip = await get_docker_container_ip(container_name)

    # connect to internal network for job communication
    process = await asyncio.create_subprocess_exec(
        "docker", "network", "connect", job_network, container_name
    )
    await process.wait()

    # wait for nginx to start
    url = f"http://{ip}/ok"
    nginx_started = await check_endpoint(url, timeout)
    if not nginx_started:
        stdout = _stdout.decode() if _stdout else ""
        stderr = _stderr.decode() if _stderr else ""
        raise Exception(f"Failed to ping nginx on {url} - server init {stdout=}, {stderr=}")


def generate_certificate(alternative_name: str) -> tuple[Certificate, RSAPrivateKey]:
    """Generate a certificate and private key"""
    private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)

    builder = x509.CertificateBuilder()
    builder = builder.subject_name(
        x509.Name([x509.NameAttribute(NameOID.COMMON_NAME, "localhost")])
    )
    builder = builder.issuer_name(x509.Name([x509.NameAttribute(NameOID.COMMON_NAME, "localhost")]))
    builder = builder.not_valid_before(datetime.now(UTC))
    builder = builder.not_valid_after(datetime.now(UTC) + timedelta(days=365))
    builder = builder.serial_number(x509.random_serial_number())
    builder = builder.public_key(private_key.public_key())
    builder = builder.add_extension(
        x509.BasicConstraints(ca=False, path_length=None), critical=True
    )

    alt_name: x509.GeneralName
    try:
        _ip = ipaddress.ip_address(alternative_name)
    except ValueError:
        alt_name = x509.DNSName(alternative_name)
    else:
        alt_name = x509.IPAddress(_ip)

    builder = builder.add_extension(
        x509.SubjectAlternativeName([x509.DNSName("localhost"), alt_name]), critical=False
    )

    certificate = builder.sign(private_key=private_key, algorithm=hashes.SHA256())

    return certificate, private_key


def serialize_certificate(certificate: Certificate) -> bytes:
    return certificate.public_bytes(encoding=serialization.Encoding.PEM)


def serialize_private_key(private_key: RSAPrivateKey) -> bytes:
    return private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )


def write_certificate(certificate: Certificate, path: Path) -> None:
    path.write_bytes(serialize_certificate(certificate))


def write_private_key(private_key: RSAPrivateKey, path: Path) -> None:
    path.write_bytes(serialize_private_key(private_key))


def save_public_key(public_key: str, dir_path: Path) -> None:
    certs_dir = dir_path / "ssl"
    client_cert_file = certs_dir / "client.crt"
    client_cert_file.write_text(public_key)


def generate_certificate_at(
    dir_path: Path | None = None, alternative_name: str = "127.0.0.1"
) -> tuple[Path, str, tuple[str, str]]:
    """
    Generate a certificate and private key and save them to a directory.
    Returns the directory path, the public key and the paths to the public and private key files.
    """

    if dir_path is None:
        dir_path = Path(tempfile.mkdtemp())
    certs_dir = dir_path / "ssl"
    certs_dir.mkdir()

    certificate, private_key = generate_certificate(alternative_name)
    public_key_path = certs_dir / "certificate.pem"
    private_key_path = certs_dir / "private_key.pem"
    write_certificate(certificate, public_key_path)
    write_private_key(private_key, private_key_path)

    public_key = serialize_certificate(certificate).decode("utf-8")
    return dir_path, public_key, (str(public_key_path), str(private_key_path))
